"""Experimental Datapipes

Collection of experimental datapipes, explicitly not ready for production,
but can be upgraded later, these might not be tested, and are to make quick experiments easier
"""
